from __future__ import annotations

import dataclasses as dc
import itertools as it
import random
import time
import typing as typ

import browser.html as html
from browser import document, bind, aio

GAME_OVER_DIV = document['minesweeper-gameover']
GAME_OVER_MESSAGE_DIV = document['minesweeper-gameover-msg']
RETRY_BUTTON = document['minesweeper-retry-btn']


@dc.dataclass(frozen=True)
class Difficulty:
    name: str
    grid_width: int
    grid_height: int
    mines_number: int
    button_color: str


DIFFICULTIES = {
    'easy': Difficulty('Easy', 9, 9, 10, 'success'),
    'normal': Difficulty('Normal', 16, 16, 40, 'primary'),
    'hard': Difficulty('Hard', 32, 16, 100, 'warning'),
    'expert': Difficulty('Expert', 32, 32, 200, 'danger'),
}

Location = typ.Tuple[int, int]


class Cell:
    def __init__(self, grid, location: Location, html_element):
        self.__grid = grid
        self.__location = location
        self.__html_element = html_element
        self.__html_element.bind('click', self._on_click)
        self.__html_element.bind('contextmenu', self._on_right_click)
        self.__is_mine = False
        self.__flagged = False
        self.__marked = False
        self.__swept = False
        self.__locked = False

    def _on_click(self, _):
        if not self.__locked and not self.__flagged:
            self.__grid.on_cell_clicked(self, True)

    def _on_right_click(self, event):
        if not self.__locked and not self.__swept:
            self.__grid.on_cell_clicked(self, False)
        event.preventDefault()

    @property
    def location(self) -> Location:
        return self.__location

    @property
    def is_mine(self) -> bool:
        return self.__is_mine

    @property
    def is_flagged(self) -> bool:
        return self.__flagged

    @property
    def is_marked(self) -> bool:
        return self.__marked

    @property
    def is_swept(self) -> bool:
        return self.__swept

    def _set_image(self, image_name, alt):
        self.__html_element.innerHTML = f'<img alt="{alt}" src="/static/minesweeper/images/tiles/{image_name}.png" />'

    def _reset_image(self):
        self.__html_element.innerHTML = ''

    def _toggle_flag(self, b: bool):
        if not self.__swept:
            self.__flagged = b
            if self.__flagged:
                self._set_image('flag', '!')
            else:
                self._reset_image()

    def _toggle_mark(self, b: bool):
        if not self.__swept:
            self.__marked = b
            if self.__marked:
                self._set_image('mark', '?')
            else:
                self._reset_image()

    def set_mine(self):
        self.__is_mine = True

    def cycle_mark(self):
        if not self.__swept:
            if self.__flagged:
                self._toggle_flag(False)
                self._toggle_mark(True)
            elif self.__marked:
                self._toggle_mark(False)
            else:
                self._toggle_flag(True)

    def sweep(self, neighbor_mines_number: int):
        if not self.__swept and not self.__flagged:
            if self.__marked:
                self._toggle_mark(False)
            if self.__is_mine:
                self.__html_element.classList.add('minesweeper-mine')
            elif neighbor_mines_number > 0:
                self._set_image(neighbor_mines_number, neighbor_mines_number)
            if not self.__is_mine:
                self.__html_element.classList.add('minesweeper-clicked')
            self.__swept = True

    def reveal_mine(self):
        if self.__is_mine:
            self._set_image('mine', 'O')
        elif self.__flagged:
            self._set_image('mine_wrong', 'X')
        self.__locked = True


# noinspection PyStatementEffect
class Grid:
    def __init__(self, game, width: int, height: int, mines_number: int):
        self.__width = width
        self.__height = height
        self.__game = game
        self.__cells: typ.List[typ.List[Cell]] = []
        self.__mines_number = mines_number
        self.__flags = 0
        self._generate_cells(width, height)

    def _generate_cells(self, width, height):
        if 'minesweeper-grid' in document:
            document['minesweeper-grid'].remove()

        table = html.TABLE(id='minesweeper-grid')
        for r in range(height):
            html_row = html.TR()
            row = []
            for c in range(width):
                html_cell = html.TD(Class='minesweeper-cell')
                row.append(Cell(self, (r, c), html_cell))
                html_row <= html_cell
            self.__cells.append(row)
            table <= html_row
        document['minesweeper-gamezone'] <= table

    @property
    def mines(self) -> int:
        return self.__mines_number

    @property
    def estimated_remaining_mines(self) -> int:
        return self.__mines_number - self.__flags

    def generate_mines(self, ignored_row: int, ignored_column: int):
        for _ in range(self.__mines_number):
            done = False
            while not done:
                r, c = random.randint(0, self.__height - 1), random.randint(0, self.__width - 1)
                if not self.__cells[r][c].is_mine and not (r == ignored_row and c == ignored_column):
                    self.__cells[r][c].set_mine()
                    done = True

    def reveal_mines(self):
        for row in self.__cells:
            for cell in row:
                cell.reveal_mine()

    def neighbors_of(self, row: int, column: int) -> (typ.List[Cell], int, int):
        nb_flagged = 0
        nb_mine = 0
        neighbors = []

        for r, c in it.product(range(row - 1, row + 2), range(column - 1, column + 2)):
            if not (r == row and c == column) and 0 <= r < self.__height and 0 <= c < self.__width:
                neighbor = self.__cells[r][c]
                nb_mine += neighbor.is_mine
                nb_flagged += neighbor.is_flagged
                neighbors.append(neighbor)

        return neighbors, nb_flagged, nb_mine

    def _sweep(self, cell: Cell, propagate: bool):
        if cell.is_mine:
            cell.sweep(0)
            return False

        neighbors, nb_flagged, nb_mines = self.neighbors_of(*cell.location)
        if not cell.is_swept:
            cell.sweep(nb_mines)

        if nb_mines == 0 or propagate and nb_flagged == nb_mines:
            for neighbor in neighbors:
                if not neighbor.is_swept and not neighbor.is_flagged:
                    self.on_cell_clicked(neighbor, True, propagate=False)

        return True

    def _check_victory(self):
        unswept = 0
        for row in self.__cells:
            for cell in row:
                # Optimization
                if cell.is_swept == cell.is_mine:
                    return False
                unswept += not cell.is_swept
        return unswept == self.__mines_number

    def on_cell_clicked(self, cell: Cell, main_click: bool, propagate: bool = True):
        if not self.__game.is_started:
            self.__game.start(cell.location)
        if main_click:
            ok = self._sweep(cell, propagate)
            if not ok:
                self.__game.stop(False)
            elif self._check_victory():
                self.__game.stop(True)
        else:
            cell.cycle_mark()
            if cell.is_flagged:
                self.__flags += 1
            elif cell.is_marked:
                self.__flags -= 1
            self.__game.update_mines_display()


class Game:
    def __init__(self, difficulty: Difficulty):
        self.__started = False
        self.__difficulty = difficulty
        self.__grid = Grid(self, difficulty.grid_width, difficulty.grid_height, difficulty.mines_number)
        self.update_mines_display()
        self.update_time_display(0)
        GAME_OVER_DIV.style.display = 'none'

    def onmessage(self, event):
        """Handles the messages sent by the timer."""
        self.update_time_display(event.data)

    @property
    def difficulty(self):
        return self.__difficulty

    @property
    def is_started(self):
        return self.__started

    def update_mines_display(self):
        document.select('#minesweeper-bombs-left .value')[0].text = str(self.__grid.estimated_remaining_mines)

    # noinspection PyMethodMayBeStatic
    def update_time_display(self, seconds: int):
        minutes = seconds // 60
        seconds = seconds % 60
        document.select('#minesweeper-timer .value')[0].text = f'{minutes:02}:{seconds:02}'

    def start(self, clicked_location: Location):
        self.__grid.generate_mines(*clicked_location)
        self.__started = True
        aio.run(self._timer())

    def stop(self, won: bool):
        self.__grid.reveal_mines()
        if won:
            GAME_OVER_DIV.classList.remove('alert-danger')
            GAME_OVER_DIV.classList.add('alert-success')
            RETRY_BUTTON.classList.remove('btn-danger')
            RETRY_BUTTON.classList.add('btn-success')
            text = 'Congratulations, you’ve won!'
            button_text = 'New Game'
        else:
            GAME_OVER_DIV.classList.remove('alert-success')
            GAME_OVER_DIV.classList.add('alert-danger')
            RETRY_BUTTON.classList.remove('btn-success')
            RETRY_BUTTON.classList.add('btn-danger')
            text = 'Game Over!'
            button_text = 'Try Again'
        GAME_OVER_MESSAGE_DIV.text = text
        RETRY_BUTTON.value = button_text
        GAME_OVER_DIV.style.display = 'block'
        self.__started = False

    async def _timer(self):
        start_time = time.time()
        while self.__started:
            t = time.time()
            await aio.sleep(0.5)
            self.update_time_display(int(t - start_time))


GAME = None


def on_new_game(event):
    global GAME
    difficulty = DIFFICULTIES[event.target.attrs['data-difficulty']]
    GAME = Game(difficulty)


@bind(document['minesweeper-retry-btn'], 'click')
def on_reset(_):
    global GAME
    GAME = Game(GAME.difficulty)


buttons_div = document['difficulty-buttons']
for key, diff in DIFFICULTIES.items():
    button = html.INPUT(type='button', Class='btn btn-' + diff.button_color, value=diff.name, data_difficulty=key)
    button.bind('click', on_new_game)
    # noinspection PyStatementEffect
    buttons_div <= button + ' '

GAME = Game(DIFFICULTIES['easy'])
